

from abc import ABC, abstractmethod
from typing import Union
from bergen.utils import ExpansionError, expandInputs
from bergen.messages.assignation import AssignationMessage
from bergen.schema import Template
from bergen.constants import OFFER_GQL, SERVE_GQL
import logging
import namegenerator
import asyncio
import websockets
from bergen.models import Node
import inspect
import sys
from aiostream import stream


logger = logging.getLogger()

import json


class BaseHelper(ABC):

    def __init__(self, peasent) -> None:
        self.peasent = peasent
        pass

    @abstractmethod
    async def pass_yield(self, message, value):
        pass

    @abstractmethod
    async def pass_progress(self, message, value):
        pass

    @abstractmethod
    async def pass_result(self, message, value):
        pass

    @abstractmethod
    async def pass_exception(self, message, exception):
        pass


class AssignationHelper():

    def __init__(self, peasent_helper: BaseHelper, assignation) -> None:
        self.peasent_helper = peasent_helper
        self.assignation = assignation
        pass


    def progress(self, value):
        self.peasent_helper.pass_progress(self.message, value)



class BasePeasent(ABC):
    ''' Is a mixin for Our Bergen '''
    helperClass = None


    def __init__(self, *args, loop = None, unique_name = None, **kwargs) -> None:
        
        
        self.unique_name = unique_name or namegenerator.gen()
        self.loop = loop or asyncio.get_event_loop()
        self.pods = {}

        self.templates_to_register = []
        self.pods_to_register = []
        self.peasent_helper = self.helperClass(self)
        super().__init__(*args, loop=loop, **kwargs)





    def register(self, node_or_template: Union[Node, Template], **kwargs):


        def real_decorator(function):
            is_coroutine = inspect.iscoroutinefunction(function)
            is_asyncgen = inspect.isasyncgenfunction(function)


            async def wrapper(message: AssignationMessage):
                
                assignation_helper = AssignationHelper(self.peasent_helper, message)
                kwargs = message.data.inputs
                
                if is_coroutine:
                    result = await function(assignation_helper, **kwargs)
                    await self.peasent_helper.pass_result(message, result)

                elif is_asyncgen:
                    yieldstream = stream.iterate(function(assignation_helper, **kwargs))
                    lastresult = None
                    async with yieldstream.stream() as streamer:
                        async for result in streamer:
                            lastresult = result
                            await self.peasent_helper.pass_yield(message, result)

                    await self.peasent_helper.pass_result(message, lastresult)
                
                else:
                    raise NotImplementedError("We do not allow for non async functions right now")
                
            
            wrapper.__name__ = function.__name__

            if isinstance(node_or_template, Node):
                logger.info(f"Registering Template for {node_or_template.name}")
                self.templates_to_register.append({
                    "node": node_or_template,
                    "params": kwargs,
                    "function": wrapper

                })

            if isinstance(node_or_template, Template):
                logger.info(f"Registering Potential Pod for {node_or_template.id}")
                self.pods_to_register.append({
                    "template": node_or_template,
                    "function": wrapper

                })


            return wrapper


        return real_decorator



    async def setup_and_run(self):
        logger.info("Offering Our Services")
        self.peasent = await SERVE_GQL.run_async(ward=self.main_ward, variables = {"name":self.unique_name})
        logger.info("Our Offer was Accesspted")


        async def offer(node: Node, params: dict, function):
            peasent_template = await OFFER_GQL.run_async(ward=self.main_ward, variables= {"node": node.id, "params": params, "peasent": self.peasent.id})
            return (peasent_template, function)

        logger.info("We are registering our functions as potential Templates")
        template_function_set = await asyncio.gather(*[offer(**values) for values in self.templates_to_register])
        self.templateid_function_map = {value[0].id: value[1] for value in template_function_set}
        self.templateid_template_map = {value[0].id: value[0] for value in template_function_set}

        logger.info(f"Following functions have been allowed! {str(self.templateid_function_map)}")

        logger.debug("Configuring")
        await self.configure()





    @abstractmethod
    async def configure(self) -> str:


        raise NotImplementedError("Please overwrite")


    def run(self):

        self.loop.create_task(self.setup_and_run())

        # we enter a never-ending loop that waits for data
        # and runs callbacks whenever necessary.
        
        self.loop.run_forever()




from abc import ABC
import re
from bergen.wards.base import BaseWard
from bergen.query import GQL, TypedGQL
import requests
from gql.transport.aiohttp import AIOHTTPTransport
import logging
from gql.transport.aiohttp import log as aio_logger
aio_logger.setLevel(logging.WARNING)
from gql import gql, Client

class GraphQlException(Exception):
    pass

class ProtocolException(GraphQlException):
    pass

class SyntaxException(GraphQlException):
    pass

class QueryException(GraphQlException):
    pass


class BaseGraphQLWard(BaseWard, ABC):
    can_subscribe = False
    

    def __init__(self, port=8000, host="localhost", protocol="ws", token=None, loop = None) -> None:
        self._graphql_endpoint = f"{protocol}://{host}:{port}/graphql"
        self._token = token
        self._headers = {"Authorization": f"Bearer: {self._token}"}
        
        self.transport =  AIOHTTPTransport(url=self._graphql_endpoint, headers=self._headers)
        self.client = Client(
        transport=self.transport, fetch_schema_from_transport=True,
        )
        
        super().__init__(loop=loop)


    async def configure(self):
        pass



    def run(self, the_query: TypedGQL, variables: dict = {}, **kwargs):
        gql_query = gql(the_query.query)
        if self.loop.is_running():
            result = self.loop.run_until_complete(self.client.execute_async(gql_query, variable_values = variables))
        else:
            result = self.client.execute(gql_query, variable_values = variables)
        return the_query.extract(result)

    async def run_async(self, the_query: TypedGQL, variables: dict = {}, **kwargs):
        
        transport =  AIOHTTPTransport(url=self._graphql_endpoint, headers=self._headers)
        client = Client(
        transport=transport, fetch_schema_from_transport=True,
        )

        gql_query = gql(the_query.query)
        result = await client.execute_async(gql_query, variable_values = variables)
        return the_query.extract(result)


import asyncio
from bergen.enums import ClientType
from bergen.wards.graphql.subscription import SubscriptionGraphQLWard
from bergen.logging import setLogging
from bergen.auths.base import BaseAuthBackend
from bergen.wards.base import BaseWard
from bergen.postmans.base import BasePostman
import logging

logger = logging.getLogger(__name__)
import os


class BaseBergen:


    def __init__(self, auth: BaseAuthBackend, host: str, port: int, protocol = "http", auto_negotiate=True, bind=True, log=logging.INFO, local=None, loop=None,  client_type: ClientType = None, **kwargs) -> None:
        
        try:
            self.loop = loop or asyncio.get_running_loop()
            self.loop_is_running = True
            logger.info("Using provided Async Loop")
        except RuntimeError:
            self.loop = asyncio.get_event_loop()
            self.loop_is_running = False
            logger.info("Created New Eventloop")

        if bind: 
            # We only import this here for typehints
            from bergen.registries.arnheim import set_current_arnheim
            set_current_arnheim(self)


        if log is not False:
            print(r"     _               _          _              ____ _ _            _    ")   
            print(r"    / \   _ __ _ __ | |__   ___(_)_ __ ___    / ___| (_) ___ _ __ | |_  ")
            print(r"   / _ \ | '__| '_ \| '_ \ / _ \ | '_ ` _ \  | |   | | |/ _ \ '_ \| __| ")
            print(r"  / ___ \| |  | | | | | | |  __/ | | | | | | | |___| | |  __/ | | | |_  ")
            print(r" /_/   \_\_|  |_| |_|_| |_|\___|_|_| |_| |_|  \____|_|_|\___|_| |_|\__| ")
            print(r"")
            setLogging(True, log)


        self.local = local if local is not None else os.getenv("ARNHEIM_LOCAL") == "1" 
        if self.local:
            logger.info("Running in Local Mode")

        self.client_type = client_type
        self.auth = auth
        self.token = self.auth.getToken()
        logger.info(" Auhorized!!!!!")

        self.main_ward = SubscriptionGraphQLWard(host=host, port=port, protocol=protocol, token=self.token, loop=self.loop)

        self._transcript = None
        self.identifierDataPointMap = {}
        self.identifierWardMap: dict[str, BaseWard] = {}


        if auto_negotiate == True and not self.loop_is_running: # If we are running in a loop we shouldn't negotiate
            self.negotiate()

        super().__init__()
            

    @property
    def transcript(self):
        assert self._transcript is not None, "We have to negotiate first with our"
        return self._transcript


    def getExtensionSettings(self, extension):
        assert extension in self.transcript.extensions, f"Arnheim seems to have no idea about this Extension {extension}"
        return self.transcript.extensions[extension]


    def getWardForIdentifier(self, identifier):
        assert self._transcript is not None, "We cannot get query Identifiers on Datapoint before having negotiated them"
        if identifier in self.identifierWardMap:
            return self.identifierWardMap[identifier]
        else:
            return self.main_ward


    def registerDatapoints(self, transcript):
        assert transcript.models is not None, "We apparently didnt't get any points"
        
        from bergen.registries.datapoint import get_datapoint_registry
        datapoint_registry = get_datapoint_registry()

        self.identifierDataPointMap = {model.identifier.lower(): model.point for model in transcript.models}
        self.identifierWardMap = {model.identifier.lower(): datapoint_registry.getClientForData(model.point, self.auth) for model in transcript.models}
        logger.info("Succesfully registered Datapoints") 



    def setPostmanFromSettings(self, transcript):
        settings = transcript.postman

        if settings.type == "pika":
            try:
                from bergen.postmans.pika import PikaPostman
                self.postman = PikaPostman(**settings.kwargs, loop=self.loop)
            except ImportError as e:
                logger.error("You cannot use the Pika Postman without installing aio_pika")
                raise e

        elif settings.type == "websocket":
            try:
                from bergen.postmans.websocket import WebsocketPostman
                self.postman = WebsocketPostman(**settings.kwargs, loop=self.loop)
            except ImportError as e:
                logger.error("You cannot use the Websocket Postman without installing websockets")
                raise e

        else:
            raise Exception(f"Postman couldn't be configured. No Postman for type {settings.type}")

        return self.postman

    
    async def negotiate_async(self, client_type=None):
        from bergen.constants import NEGOTIATION_GQL
        
        # We resort escalating to the different client Type protocols
        clientType = client_type or self.client_type or self.auth.getClientType()
        self._transcript = await NEGOTIATION_GQL.run_async(ward=self.main_ward, variables={"clientType": clientType, "local": self.local})
        self.registerTranscript(self._transcript)

        postman = self.setPostmanFromSettings(self._transcript)
        await postman.configure()

    async def disconnect_async(self, client_type=None):
        await self.postman.disconnect()



    def registerTranscript(self, transcript):
        logger.info(f"Successfully Got Protocols")

        if self._transcript:
            self.registerDatapoints(self._transcript)

        logger.info("We established all Proper Clients for the interaction!")
        logger.info("Arnheim Ready!!!!!!!!")



    def negotiate(self, client_type = None):
        from bergen.constants import NEGOTIATION_GQL
        
        # We resort escalating to the different client Type protocols
        clientType = client_type or self.client_type or self.auth.getClientType()
        self._transcript = NEGOTIATION_GQL.run(ward=self.main_ward, variables={"clientType": clientType or self.auth.getClientType(), "local": self.local})
        self.registerTranscript(self._transcript)

        postman = self.setPostmanFromSettings(self._transcript)
        self.loop.run_until_complete(postman.configure())
        


    def getWard(self) -> BaseWard:
        return self.main_ward

    def getPostman(self) -> BasePostman:
        return self.postman

    def _repr_html_(self):
        return f"""
            <p> Arnheim Client <p>
            <table>
                <tr>
                    <td> Connected to </td> <td> {self.main_ward.host} </td>
                </tr>
            </table>



        """


    async def __aenter__(self):
        await self.negotiate_async()
        print(self)


    async def __aexit__(self,*args, **kwargs):
        await self.disconnect_async()




from bergen.wards.graphql.base import QueryException
from bergen.query import TypedGQL
import json
from bergen.wards.graphql.default import GraphQlWard
import logging
import websockets
import uuid

logger = logging.getLogger(__name__)

class SubscriptionGraphQLWard(GraphQlWard):
    subprotocols = ["graphql-ws"]
    extra_headers = {}
    can_subscribe = True

    def __init__(self, port=8000, host="localhost", protocol="http", token=None, loop=None) -> None:
        self._websocket_endpoint = f"ws://{host}:{port}/graphql?token={token}" 
        super().__init__(port=port, host=host, protocol=protocol, token=token, loop=loop)




from bergen.schema import Node
from bergen.types.node.inputs import Inputs, Outputs
from bergen.managers.model import ModelManager
from bergen.queries.delayed.node import NODE_QUERY, CREATE_NODE_MUTATION
from typing import Type


class NodeManager(ModelManager[Node]):

    def get_or_create(self, inputs: Type[Inputs] = None, outputs: Type[Outputs] = None , **kwargs) -> Node:
        
        parsed_inputs = inputs.serialized
        parsed_outputs = outputs.serialized
        
        node = CREATE_NODE_MUTATION(self.model).run(variables={
            "inputs" : parsed_inputs,
            "outputs": parsed_outputs,
            **kwargs

        })
        return node


from typing import Generic
from bergen.query import  TypedGQL
from typing import TypeVar


T = TypeVar("T")

class BaseWard(Generic[T]):

    def __init__(self, loop=None):
        self.loop = loop
        if self.loop.is_running():
            self.loop.create_task(self.configure())
        else:
            self.loop.run_until_complete(self.configure())


    def run(self, gql: TypedGQL, variables: dict = {}):
        return gql.cls(**{})


from typing import Any
from bergen.schema import AssignationParams, Template
from bergen.wards.graphql.subscription import SubscriptionGraphQLWard
from bergen.registries.arnheim import get_current_arnheim
import asyncio
from bergen.types.model import ArnheimModel
from bergen.extenders.base import BaseExtender
import logging
from aiostream import stream
import uuid



logger = logging.getLogger(__name__)



def serialize(inputs: dict):
    serialized_inputs = {}
    
    for key, value in inputs.items():
        if isinstance(value, ArnheimModel):
            serialized_inputs[key] = value.id
        else:
            serialized_inputs[key] = value

    return serialized_inputs


class AsyncPodHandler:

    def __init__(self, provisionHandler, ward, node, pod) -> None:
        self._pod = pod
        self._node = node
        self._provisionHandler = provisionHandler
        self._ward = ward
        super().__init__()


    async def stream(self, inputs: dict):
        from arnheim.constants import ASSIGN_GQL
        logger.info(f"Trying to assign to {self._pod}")
        serialized_inputs = {}

        serialized_inputs = serialize(inputs)

        expandedoutputs = None
        async for item in ASSIGN_GQL.subscribe(self._ward, variables={"pod": self._pod.id, "inputs": serialized_inputs, "reference": str(uuid.uuid4())}):
            if item.status == "DONE":
                from arnheim.utils import expandOutputs
                expandedoutputs = expandOutputs(self._node, item.outputs)
                yield expandedoutputs
                break
            elif item.status == "YIELD":
                from arnheim.utils import expandOutputs
                expandedoutputs = expandOutputs(self._node, item.outputs)
                yield expandedoutputs
            elif item.status == "CRITICAL":
                raise Exception(item.statusmessage)
            else:
                pass


    async def assign(self, inputs: dict):
        from arnheim.constants import ASSIGN_GQL
        logger.debug(f"Trying to assign to {self._pod}")
        serialized_inputs = {}

        serialized_inputs = serialize(inputs)

        async for item in ASSIGN_GQL.subscribe(self._ward, variables={"pod": self._pod.id, "inputs": serialized_inputs, "reference": str(uuid.uuid4())}):
            logger.info(f"Received {item}")
            if item.status == "DONE":
                from arnheim.utils import expandOutputs
                expandedoutputs = expandOutputs(self._node, item.outputs)
                return expandedoutputs
            elif item.status == "CRITICAL":
                raise Exception(item.statusmessage)
            else:
                pass


    async def kill(self):
        await asyncio.sleep(0.01)
        logger.info("Closed Connection")


class ProvisionHandler:

    def __init__(self, nodeModel, ward: SubscriptionGraphQLWard, template: Template = None) -> None:
        super().__init__()
        self._pod = None
        self._ward = ward
        self._template = template
        self._node = nodeModel

    async def __aenter__(self):
        from bergen.constants import PROVIDE_GQL

        async for item in PROVIDE_GQL.subscribe(self._ward, variables={"node": self._node.id, "selector": {"template": self.template.id, "kwargs": {"volunteer": 2}}, "reference": str(uuid.uuid4())}):
            if item.pod.status == "ACTIVE":
                self._provisionreference = item.reference 
                self._pod = AsyncPodHandler(self, self._ward, self._node, item.pod)
                return self._pod
            else:
                pass

    


    async def assign(self, inputs):
        async with self as handler:
            return await handler.assign(inputs)

        
    async def __aexit__(self, exc_type, exc, tb):
        await self._pod.kill()
        print(exc_type, exc, tb)
        self._pod = None





class NodeExtender(BaseExtender):


    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args,**kwargs)
        
        self._postman = get_current_arnheim().getPostman()

    async def __call_async(self, inputs: dict, params: AssignationParams, **kwargs):
        
        return await self._postman.assign(self, inputs, params, **kwargs)

    async def __delay_async(self, inputs: dict, params: AssignationParams, **kwargs):
    
        return await self._postman.delay(self, inputs, params, **kwargs)

    def stream(self, inputs: dict, params: AssignationParams = None, **kwargs):

        return stream.iterate(self._postman.stream(self, inputs, params, **kwargs))
    
    def delay(self, inputs: dict, params: AssignationParams = None, **kwargs):
        try:
            event_loop = asyncio.get_event_loop()
        except RuntimeError:
            pass
        else:
            if event_loop.is_running():
                return self.__delay_async(inputs, params, **kwargs)
            else:
                logger.info("Spinning up new eventloop just for that <3")
                event_loop = asyncio.get_event_loop()
                result = event_loop.run_until_complete(self.__delay_async(inputs, params, **kwargs))
                return result



    def __call__(self, inputs: dict, params: AssignationParams = None, **kwargs) -> dict:
        """Call this node (can be run both asynchronously and syncrhounsly)

        Args:
            inputs (dict): The inputs for this Node
            params (AssignationParams, optional): [description]. Defaults to None.

        Returns:
            outputs (dict): The ooutputs of this Node
        """
        try:
            event_loop = asyncio.get_event_loop()
        except RuntimeError:
            pass
        else:
            if event_loop.is_running():
                return self.__call_async(inputs, params, **kwargs)
            else:
                logger.info("Spinning up new eventloop just for that <3")
                event_loop = asyncio.get_event_loop()
                result = event_loop.run_until_complete(self.__call_async(inputs, params, **kwargs))
                return result

    def provide(self, provider = "vart"):

        ward = get_current_arnheim().getWard()
        self._provisionhandler = ProvisionHandler(self, ward, provider=provider)
        return self._provisionhandler


    def _repr_html_(self):
        string = f"{self.name}</br>"

        for input in self.inputs:
            string += "Inputs </br>"
            string += f"Port: {input._repr_html_()} </br>"

        for output in self.outputs:
            string += "Outputs </br>"
            string += f"Port: {output._repr_html_()} </br>"


        return string


from bergen.enums import GrantType
from bergen.clients.mixins.hostmixin import HostMixIn
from bergen.wards.graphql.subscription import SubscriptionGraphQLWard
from bergen.auths.backend.app import ArnheimBackendOauth
from bergen.clients.base import BaseBergen
from bergen.clients.mixins.querymixin import QueryMixIn
from bergen.clients.mixins.subscribemixin import SubscribeMixIn


class BergenHost(BaseBergen, QueryMixIn, SubscribeMixIn, HostMixIn ):

    def __init__(self, host: str = "localhost", port: int = 8000, client_id: str = None, client_secret: str = None, protocol="http", bind=True, log=True, name="", grant_type: GrantType = GrantType.BACKEND) -> None:

        auth = ArnheimBackendOauth(host=host, port=port, client_id=client_id, client_secret=client_secret, protocol="http", verify=True)

        main_ward = SubscriptionGraphQLWard(host=host, port=port, protocol=protocol, token=auth.getToken())

        super().__init__(auth, main_ward, auto_negotiate=True, bind=bind, log=log)


from bergen.messages.types import ASSIGNATION_REQUEST
from bergen.messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from pydantic import BaseModel
from typing import Optional


class AssignationRequestParams(BaseModel):
    pass

class AssignationRequestMetaAuthModel(MessageMetaExtensionsModel):
    token: str

class AssignationRequestMetaExtensionsModel(MessageMetaExtensionsModel):
    progress: Optional[str]

class AssignationRequestMetaModel(MessageMetaModel):
    type: str = ASSIGNATION_REQUEST
    auth: AssignationRequestMetaAuthModel
    extensions: Optional[AssignationRequestMetaExtensionsModel]

class AssignationRequestDataModel(MessageDataModel):
    node: Optional[int] #TODO: Maybe not optional
    pod: Optional[int]
    template: Optional[int]
    reference: str
    callback: Optional[str]
    progress: Optional[str]

    inputs: dict
    params: Optional[AssignationRequestParams]


class AssignationRequestMessage(MessageModel):
    data: AssignationRequestDataModel
    meta: AssignationRequestMetaModel